# A SAT Instance Generator to Analyse Multiwinner Voting Rules with Approval Ballots

This is the supplementary material belonging to the paper "Analysing Irresolute Multiwinner Voting Rules with Approval Ballots via SAT Solving" by Boas Kluiving, Adriaan de Vries, Pepijn Vrijbergen, Arthur Boixel, and Ulle Endriss, published in the Proceedings of the 24th European Conference on Artificial Intelligence (ECAI-2020). In the paper, we establish various results regarding the possibility and impossibility of designing multiwinner voting rules that satisfy certain axioms. This repository contains the code used to generate the SAT instances that correspond to some of these results.


## Usage

The program generates, given a list of axioms, the propositional formulas representing these axioms in conjunctive normal form (CNF) and writes them to a file in DIMACS format. We can then check whether these axioms are satisfiable or not using any SAT solver, such as Lingeling or MiniSAT. We might even run tools that can find minimal unsatisfiable subsets of these sets of clauses, such as MUSer2. Using the '-s' command-line option, the program can also directly check satisfiability after generating the clauses (see dependencies).

Note that this program will by default generate SAT instances for the scenario with 3 voters (n=3), 4 alternatives (m=4), and committee size 3 (k=3). These parameters can be changed by changing the corresponding parameters in the file 'config.py'.

```
usage: python3 main.py [-h] [-l] [-o OUTPUTFILE] [-s] [-p] axioms [axioms ...]

positional arguments:
  axioms                the list of axioms, separated by spaces, to include in the CNF

optional arguments:
  -h, --help            show this help message and exit
  -l, --list            print a list of all implemented axioms and exit
  -o OUTPUTFILE, --outputfile OUTPUTFILE
                        the name of the file to which the CNF will be written (default: 'output.dimacs')
  -s, --satsolve        check whether the generated CNF is satisfiable (note: requires 'pylgl')
  -p, --progressbar     show a progress bar during the generation of the axioms (note: requires 'tqdm')
```


## Implemented Axioms

The following axioms are available for use in our program:
* CardinalityParetoEfficiency
* CautiousCardinalityAddingStrategyproofness
* CautiousCardinalityDroppingStrategyproofness
* CautiousCardinalityStrategyproofness
* CautiousSupersetStrategyproofness
* ExtendedJustifiedRepresentation
* JustifiedRepresentation
* Lemma1Proportionality
* MinimalProportionality
* OptimisticCardinalityStrategyproofness
* OptimisticSupersetStrategyproofness
* PAVVotingRule
* PessimisticCardinalityStrategyproofness
* PessimisticSupersetStrategyproofness
* Resolute
* ResoluteDroppingStrategyproofness
* SupersetParetoEfficiency
* WeakEfficiency

Note: Technically, "AtLeastOne" (expressing that a voting rule should in every situation always choose at least one committee) is also an axiom, but this axiom is always added automatically.


## Example

As an example, suppose you would like to check whether ExtendedJustifiedRepresentation and PessimisticCardinalityStrategyproofness are compatible for n=3, m=4, and k=3. We can use the following command for this:

`python3 main.py ExtendedJustifiedRepresentation PessimisticCardinalityStrategyproofness`

The resulting clauses will by default be written to the file 'output.dimacs' in the same directory. If you would like to choose a different file name, use the '-o' option. Using the '-p' option furthermore shows a progress bar for the generation of the axioms:

`python3 main.py ExtendedJustifiedRepresentation PessimisticCardinalityStrategyproofness -o customFile.dimacs -p`

Finally, if you would like to run the SAT solver directly on the generated set of clauses, use the '-s' option (the clauses will still also be output as a DIMACS file):

`python3 main.py ExtendedJustifiedRepresentation PessimisticCardinalityStrategyproofness -s`


## Dependencies

Python 3 is required to run the program. It has been tested using Python 3.8.1 but may also work under earlier/later versions.
In principle, there are no additional dependencies. However, to be able to show a progress bar (option '-p') or to check the satisfiability of the clauses generated directly (option '-s'), the following two modules need to be installed:

1. tqdm (pip install tqdm)
2. Lingeling (pip install pylgl)

## Reproducing the Results of the Paper

**Lemma 2.**
Our main impossibility result concerns voting rules that are minimally proportional, (cardinality) Pareto efficient, and (cardinality) immune to manipulation by cautious voters. This result can be verified for the case of n=3, m=4, and k=3 by using the following command:

`python3 main.py CautiousCardinalityStrategyproofness MinimalProportionality CardinalityParetoEfficiency -o lemma2.dimacs -s`

Answer: `Unsatisfiable`

**Proposition 3.**
We obtain a possibility result when we drop the Pareto efficiency axiom:

`python3 main.py CautiousCardinalityStrategyproofness MinimalProportionality -o prop3.dimacs -s`

Answer: `Satisfiable`

**Varying strategyproofness (dropping candidates).**
When we impose the restriction that voters will only manipulate by dropping candidates from their ballots, the impossibility still remains:

`python3 main.py CautiousDroppingCardinalityStrategyproofness MinimalProportionality CardinalityParetoEfficiency -o stratproof_drop.dimacs -s`

Answer: `Unsatisfiable`

**Varying strategyproofness (adding candidates).**
If, on the other hand, voters can only manipulate by adding candidates to their ballot, the result becomes a possibility. Moreover, Proportional Approval Voting (PAV) actually satisfies all these properties, which can be seen by executing the following command (the "axiom" PAVVotingRule simply requires agreement with PAV):

`python3 main.py CautiousAddingCardinalityStrategyproofness MinimalProportionality CardinalityParetoEfficiency PAVVotingRule -o stratproof_add.dimacs -s`

Answer: `Satisfiable`

**Varying strategyproofness (optimistic preference extension).**
If we replace strategyproofness for cautious voters by strategyproofness for optimistic voters, then we obtain a possibility:

`python3 main.py OptimisticCardinalityStrategyproofness MinimalProportionality CardinalityParetoEfficiency -o stratproof_opt.dimacs -s`

Answer: `Satisfiable`

**Varying strategyproofness (pessimistic preference extension).**
If we replace strategyproofness for cautious voters by strategyproofness for pessimistic voters, then we again obtain a possibility:

`python3 main.py PessimisticCardinalityStrategyproofness MinimalProportionality CardinalityParetoEfficiency -o stratproof_pess1.dimacs -s`

Answer: `Satisfiable`

**Varying strategyproofness (pessimistic preference extension, stronger form of proportionality).**
If we however also strengthen minimal proportionality to the proportionality axiom of Lemma 1 in the paper, then we obtain an impossibility:

`python3 main.py PessimisticCardinalityStrategyproofness Lemma1Proportionality CardinalityParetoEfficiency -o stratproof_pess2.dimacs -s`

Answer: `Unsatisfiable`

**Varying efficiency.**
If we replace cardinality Pareto efficiency by the weak efficiency axiom proposed by Dominik Peters, then we obtain a possibility:

`python3 main.py CautiousCardinalityStrategyproofness MinimalProportionality WeakEfficiency -o weak_efficiency.dimacs -s`

Answer: `Satisfiable`

**Varying assumptions on preferences.**
If we replace all cardinality axioms by their superset counterparts, then the original impossibility becomes a possibility:

`python3 main.py CautiousSupersetStrategyproofness MinimalProportionality SupersetParetoEfficiency -o superset.dimacs -s`

Answer: `Satisfiable`
