##########
# This code is supplementary material belonging to the paper "Analysing 
# Irresolute Multiwinner Voting Rules with Approval Ballots via SAT Solving",
# published in ECAI-2020.
##########
# This program generates SAT clauses in CNF in DIMACS format that express
# certain desirable properties of multiwinner approval voting rules.
# SAT-solving these generated clauses then allows you to either find a
# suitable voting rule or an impossibility result of these properties.
##########
# Authors: Adriaan de Vries & Boas Kluiving
##########
# File: helperfunctions.py
# This file contains some useful helper functions that are used 
# by the axiom clause-generation functions in main.py
##########

from __future__ import division
from math import factorial, ceil
from fractions import Fraction
from itertools import product, combinations, permutations, chain
from sys import stderr
from copy import deepcopy
import config

# Parameters: these are imported from config.py and should be set there
n = config.n
m = config.m
k = config.k

# This sets the starting and end value, for if you want more than 1 value of k. m-1 and m give only k=m-1

# This is not currently used, but having more than one value for k is necessary for an investigation into
# certain axioms, such as committee monotonicity. The currently implemented axioms have not been well-tested
# with multiple values.
k0 = k
k1 = k+1
ks = range(k0, k1)


# these functions are necessary for the creation of the cnf of the axioms
 
def allVoters():
    """Returns all voters, represented as integers"""

    return range(n)

def allAlternatives():
    """Returns all alternatives, represented as integers"""

    return range(m)

def allBallots():
    """Returns all ballots, as binary tuples"""

    r = product(*([[0,1]]*m))
    return r

def allProfiles():
    """Returns all profiles, as iterables of ballots"""

    return product(*([list(allBallots())]*n))

def allCommittees():
    """Returns all possible committees as binary tuples, given possible values of k"""

    for b in allBallots():
        if sum(b) in ks:
            yield b

def allCommitteesOfSize(k):
    """Returns all possible committees of given exact size"""

    for c in combinations(range(m), k):
        com = [0]*m
        for i in c:
            com[i] = 1
        yield tuple(com)

def allWinningSets():
    """Returns all winning sets, as a generator"""

    s = list(allCommittees())
    # from itertools recipe for powerset
    for W in chain.from_iterable(combinations(s, r) for r in range(len(s)+1)):
        #only non-empty winning sets are allowed
        if W:
            yield W

def union(p):
    """Given binary tuples, computes OR which represents set union"""

    union = []
    for thing in zip(*p):
        if 1 in thing:
            union.append(1)
        else:
            union.append(0)
    return tuple(union)

def singletonBallot(a):
    """Returns the singleton ballot {a}"""

    ballot = [0]*m
    ballot[a] = 1
    return tuple(ballot)

def ivariants(voter, profile):
    """Given a ballot profile and a voter, returns all profiles which differ only in that voter"""

    ivar = list(profile)
    for ballot in allBallots():
        result = deepcopy(ivar)
        result[voter] = ballot
        yield tuple(result)

def isSubsetOf(ballot1, ballot2, strict=True):
    """Checks whether ballot1 is a (strict) subset of ballot2"""

    for b1,b2 in zip(ballot1, ballot2):
        if b1 and not b2:
            return False
        if b2 and not b1:
            strict = False
    
    if strict:
        return False
    else:
        return True

def intersection(committee, ballot):
    """Given binary tuples, computes AND which represents set intersection"""

    result = []

    for c,v in zip(committee, ballot):
        if c and v:
            result.append(1)
        else:
            result.append(0)
    return tuple(result)

def strictlyBetter(voter, committee1, committee2, profile):
    """Checks if committee1 is strictly superset-better than committee2 for voter in profile"""

    successes1 = intersection(committee1, profile[voter])
    successes2 = intersection(committee2, profile[voter])

    return intersection(successes1, successes2) == successes1 and successes2 != successes1

def weaklyBetter(voter, committee1, committee2, profile):
    """Checks if committee1 is at least as superset-good as committee2 for voter in profile"""

    successes1 = intersection(committee1, profile[voter])
    successes2 = intersection(committee2, profile[voter])

    return intersection(successes1, successes2) == successes1

def isPartylistProfile(profile):
    """
    Checks if profile is a party-list profile.
    A profile is a party-list profile iff, after removing duplicate ballots,
    there is no more overlap in the ballots
    """

    unique_ballots = set(profile)
    for e in zip(*unique_ballots):
        if sum(e)>1:
            return False
    return True

def isSemiPartylistProfile(profile, a):
    """
    Checks if profile is a semi-partylist profile, 
    in which voters either approve only 'a' or don't approve 'a'
    """

    for ballot in profile:
        if not (ballot[a] == 0 or (sum(ballot) == 1 and ballot[a] == 1)):
            return False
    return True

def PAVscore(profile, committee):
    """
    Returns the PAV score of a committee given a profile
    For very, very large k (>300) this function does not work well enough
    because of floating point errors
    """
    
    score = 0
    for voter in profile:
        approvecount = 0
        for elected,vote in zip(committee, voter):
            if vote and elected:
                approvecount+=1
        for i in range(approvecount):
            score+=(1/(i+1))
    return score

def cardinalityOfOverlap(ballot, committee):
    """Gives the amount of approved candidates in committee according to ballot"""

    c=0
    for candidate, vote in zip(committee, ballot):
        c += (candidate and vote) 
    return c


def strictlySupersetDominates(profile, committee1, committee2):
    """Returns true if committee1 strictly superset-dominates committee"""

    strict = False
    for voter, ballot in enumerate(profile):
        int1 = intersection(ballot, committee1)
        int2 = intersection(ballot, committee2)
        for i1, i2 in zip(int1, int2):
            if i2 > i1:
                return False
            if i1 > i2:
                strict = True
    return strict

def satisfiesJR(committee, profile, debug=False):
    """Return true if the committee, given the profile, satisfies JR"""

    k = sum(committee)
    threshold = n/k - 1e-5
    uncovered_ballots = []
    for ballot in profile:
        for vote, elected in zip(ballot, committee):
            if vote and elected:
                break
        else:
            uncovered_ballots.append(ballot)
    votespercandidate = [0]*m
    for ballot in uncovered_ballots:
        for candidate, vote in enumerate(ballot):
            votespercandidate[candidate]+=vote
    
    if max(votespercandidate) >= threshold:
        return False 
    else:
        return True
        
def satisfiesEJR(committee, profile):
    """Return true if the committee, given the profile, satisfies EJR"""

    k = sum(committee)
    
    # calculate for each voter how many representatives he has in committee
    voterRepr = []
    for ballot in profile:
        representatives = 0
        for c,b in zip(committee, ballot):
            if c and b:
                representatives += 1
        voterRepr.append(representatives)

    for l in range(1, k+1):
        # calculate all voters with < l representatives
        underRepres = []
        for voter, repres in enumerate(voterRepr):
            if repres < l:
                underRepres.append(voter)

        minSize = ceil(l * (n/k) - 1e-5)

        # find all subsets of size l * (n/k)
        for comb in combinations(underRepres, minSize):
            
            # count how many alternatives they all agree on
            approveCount = 0
            for i in range(m):
                allApprove = True
                for voter in comb:
                    if profile[voter][i] != 1:
                        allApprove = False
                if allApprove:
                    approveCount += 1

            # this should not be >= l if it satisfies EJR
            if approveCount >= l:
                return False

    return True

def strictlyDominates(profile, committee1, committee2):
    """
    Checks if every voter has at least as many of their candidates in
    committee1 as in committee2, and at least one has strictly more
    """

    strict = False
    for ballot in profile:
        card1 = cardinalityOfOverlap(ballot, committee1)
        card2 = cardinalityOfOverlap(ballot, committee2)
        if card1 < card2:
            return False
        elif card1 > card2:
            strict = True
    return strict


def minCardinality(winning_set, ballot):
    """
    Gets the minimum utility defined as amount of approved candidates in 
    a committee from the winning set
    """

    return min([cardinalityOfOverlap(ballot, committee) for committee in winning_set])

def maxCardinality(winning_set, ballot):
    """
    Gets the maximum utility defined as amount of approved candidates in 
    a committee from the winning set
    """

    return max([cardinalityOfOverlap(ballot, committee) for committee in winning_set])
