##########
# This code is supplementary material belonging to the paper "Analysing 
# Irresolute Multiwinner Voting Rules with Approval Ballots via SAT Solving",
# published in ECAI-2020.
##########
# This program generates SAT clauses in CNF in DIMACS format that express
# certain desirable properties of multiwinner approval voting rules.
# SAT-solving these generated clauses then allows you to either find a
# suitable voting rule or an impossibility result of these properties.
##########
# Authors: Adriaan de Vries & Boas Kluiving
##########
# File: main.py
# This file contains the functions to generate CNF clauses for
# axioms, as well as the command-line interface to run the program.
##########


from itertools import combinations
from sys import stderr
from helperfunctions import *
from dimacser import dimacs

import argparse

SHOW_PROGRESS = False


def progressbar(a):
    if SHOW_PROGRESS:
        return tqdm.tqdm(a)
    else:
        return a


# necessary for converting python objects to ints, which are 
# required for standardized SAT formats.
litcount = 0
lit2int = {}
int2lit = {}

# checks if a literal has already been defined for given committee
# and profile, and returns it (making a new one if necessary).
def posLiteral(committee, profile):
    global litcount, lit2int, int2lit

    if (committee, profile) in lit2int:
        return lit2int[(committee, profile)]
    else:
        litcount+=1
        lit2int[(committee, profile)] = litcount
        int2lit[litcount] = (committee, profile)
        return litcount

def negLiteral(committee, profile):
    return -posLiteral(committee, profile)



def cnfPAVVotingRule():
    """
    Calculates the PAV score, and returns a negative literal for those
    committees that do not maximize it.

    Note that this does not fully specify a voting rule, as ties are 
    allowed to be broken in any way, including not at all.
    """


    cnf  = []
    for p in progressbar(list(allProfiles())):
        for k in ks:
            comscore = []
            # get all scores
            for c in allCommitteesOfSize(k):
                score = PAVscore(p,c)
                comscore.append((c, score))
            # calculate the maximum score
            maxscore = 0
            for c,s in comscore:
                maxscore = max(maxscore, s)
            # add all suboptimal committees as neglits
            for c,s in comscore:
                if s + 1e-5 < maxscore:
                    cnf.append([negLiteral(c,p)])
    return cnf

def cnfResoluteDroppingStrategyproofness():
    """
    This axiom currently works according to "Strategyproofness" in D. Peters' 
    paper, and therefore assumes resoluteness. States that voters should not be 
    able to get a better outcome by approving a subset of their truthfully 
    approved candidates.
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():

            for p2 in ivariants(i, p1):
                if isSubsetOf(p2[i], p1[i]):
                    for c1 in allCommittees():
                        k = sum(c1)
                        for c2 in allCommitteesOfSize(k):
                            if strictlyBetter(i, c1, c2, p1):
                                cnf.append([negLiteral(c1, p1), negLiteral(c2, p2)])
    return cnf

def cnfCautiousCardinalityStrategyproofness():
    """
    Kelly-strategyproofness. A winning set under this
    definition is better if every committee in it is better.
    Better, for committees, is defined as the cardinality
    of the overlap between the ballot and the committee.
    
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                for X in allWinningSets():
                    for Y in allWinningSets():
                        if minCardinality(Y, p1[i]) >= maxCardinality(X, p1[i]) and maxCardinality(Y, p1[i]) > minCardinality(X, p1[i]):
                            X_is_not_winning_set_for_p1 = [negLiteral(c, p1) if c in X else posLiteral(c, p1) for c in allCommittees()]
                            Y_is_not_winning_set_for_p2 = [negLiteral(c, p2) if c in Y else posLiteral(c, p2) for c in allCommittees()]
                            cnf.append(X_is_not_winning_set_for_p1 + Y_is_not_winning_set_for_p2)
    return cnf

def cnfCautiousCardinalityDroppingStrategyproofness():
    """
    Same as CautiousCardinalityStrategyproofness,
    except we only consider manipulations in which
    candidates were dropped from the truthful ballot.
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                # the ivariant must be a subset
                if not isSubsetOf(p2[i], p1[i]):
                    continue
                for X in allWinningSets():
                    for Y in allWinningSets():
                        if minCardinality(Y, p1[i]) >= maxCardinality(X, p1[i]) and maxCardinality(Y, p1[i]) > minCardinality(X, p1[i]):
                            X_is_not_winning_set_for_p1 = [negLiteral(c, p1) if c in X else posLiteral(c, p1) for c in allCommittees()]
                            Y_is_not_winning_set_for_p2 = [negLiteral(c, p2) if c in Y else posLiteral(c, p2) for c in allCommittees()]
                            cnf.append(X_is_not_winning_set_for_p1 + Y_is_not_winning_set_for_p2)
    return cnf

def cnfCautiousCardinalityAddingStrategyproofness():
    """
    Same as CautiousCardinalityStrategyproofness,
    except we only consider manipulations in which
    candidates were added to the truthful ballot.
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                # the original ballot must be a subset of the i-variant
                if not isSubsetOf(p1[i], p2[i]):
                    continue
                for X in allWinningSets():
                    for Y in allWinningSets():
                        if minCardinality(Y, p1[i]) >= maxCardinality(X, p1[i]) and maxCardinality(Y, p1[i]) > minCardinality(X, p1[i]):
                            X_is_not_winning_set_for_p1 = [negLiteral(c, p1) if c in X else posLiteral(c, p1) for c in allCommittees()]
                            Y_is_not_winning_set_for_p2 = [negLiteral(c, p2) if c in Y else posLiteral(c, p2) for c in allCommittees()]
                            cnf.append(X_is_not_winning_set_for_p1 + Y_is_not_winning_set_for_p2)
    return cnf


def cnfCautiousSupersetStrategyproofness():
    """
    This axiom also implements Kelly-strategyproofness,
    with one committee being better than another if the 
    overlap between the committee and a ballot is a superset.
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                for X in allWinningSets():
                    for Y in allWinningSets():
                        is_better = True
                        is_same = True
                        for y in Y:
                            for x in X:
                                if not weaklyBetter(i,x,y,p1):
                                    is_better = False
                                if not weaklyBetter(i,y,x,p1):
                                    is_same = False

                        if is_better and not is_same:
                            X_is_not_winning_set_for_p1 = [negLiteral(c, p1) if c in X else posLiteral(c, p1) for c in allCommittees()]
                            Y_is_not_winning_set_for_p2 = [negLiteral(c, p2) if c in Y else posLiteral(c, p2) for c in allCommittees()]
                            cnf.append(X_is_not_winning_set_for_p1 + Y_is_not_winning_set_for_p2)
    return cnf


def cnfOptimisticCardinalityStrategyproofness():
    """
    Cardinality-based strategyproofness with respect to optimistic voters
    """
    cnf  = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                for c2 in allCommittees():
                    k = sum(c2)
                    clause = [negLiteral(c2, p2)]
                    card = cardinalityOfOverlap(p1[i], c2)
                    for c1 in allCommitteesOfSize(k):
                        if cardinalityOfOverlap(p1[i], c1) >= card:
                            clause.append(posLiteral(c1, p1))
                    cnf.append(clause)
    return cnf

def cnfPessimisticCardinalityStrategyproofness():
    """
    Cardinality-based strategyproofness with respect to pessimistic voters
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i,p1):
                for c1 in allCommittees():
                    k = sum(c1)
                    clause = [negLiteral(c1, p1)]
                    card = cardinalityOfOverlap(p1[i], c1)
                    for c2 in allCommitteesOfSize(k):
                        if cardinalityOfOverlap(p1[i], c2) <= card:
                            clause.append(posLiteral(c2, p2))
                    cnf.append(clause)
    return cnf

def cnfOptimisticSupersetStrategyproofness():
    """
    Superset-strategyproofness with respect to optimistic voters
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i, p1):
                for c2 in allCommittees():
                    k=sum(c2)
                    clause = [negLiteral(c2, p2)]
                    for c1 in allCommitteesOfSize(k):
                        if not strictlyBetter(i, c1, c2, p1):
                            clause.append(posLiteral(c1, p1))
                    cnf.append(clause)
    return cnf

def cnfPessimisticSupersetStrategyproofness():
    """
    Superset-strategyproofness with respect to pessimistic voters
    """
    cnf = []
    for p1 in progressbar(list(allProfiles())):
        for i in allVoters():
            for p2 in ivariants(i, p1):
                for c1 in allCommittees():
                    k = sum(c1)
                    clause = [negLiteral(c1, p1)]
                    for c2 in allCommitteesOfSize(k):
                        if not strictlyBetter(i, c1, c2, p1):
                            clause.append(posLiteral(c2, p2))
                    cnf.append(clause)
    return cnf

def cnfSupersetParetoEfficiency():
    """
    Superset-based Pareto efficiency.
    
    Don't return a committee if there is another committee where every
    voter gets a superset of their approved candidates in the committee,
    at least one of which is strict.
    """
    cnf=[]
    for profile in progressbar(list(allProfiles())):
        for k in ks:
            for committee1 in allCommitteesOfSize(k):
                for committee2 in allCommitteesOfSize(k):
                    if strictlySupersetDominates(profile, committee1, committee2):
                        cnf.append([negLiteral(committee2, profile)])
    return cnf


def cnfCardinalityParetoEfficiency():
    """
    Cardinality-based Pareto efficiency.
    
    Don't return a committee if there is another committee where every
    voter gets at least as many of their approved candidates in the 
    committee, and at least one gets strictly more.
    """
    cnf=[]
    for profile in progressbar(list(allProfiles())):
        for k in ks:
            for committee1 in allCommitteesOfSize(k):
                for committee2 in allCommitteesOfSize(k):
                    if strictlyDominates(profile, committee1, committee2):
                        cnf.append([negLiteral(committee2, profile)])
    return cnf

def cnfMinimalProportionality():
    """
    Corresponds to Peters' weakest proportionality axiom, which he calls 
    "proportionality", which means that if a singleton ballot for candidate 
    'a' appears at least n/k times in a party-list profile, 'a' should be 
    elected.
    
    In the irresolute case, states that all winners should satisfy
    proportionality.
    """
    cnf = []
    for p in progressbar(list(allProfiles())):
        if isPartylistProfile(p):
            for a in allAlternatives():
                for k in ks:
                    # the constant is there to prevent floating-point errors
                    if p.count(singletonBallot(a)) >= n/k-1e-5:
                        for c in allCommitteesOfSize(k):
                            if not c[a]:
                                cnf.append([negLiteral(c, p)])
    return cnf


def cnfLemma1Proportionality():
    """
    Corresponds to a strengthening of minimal proportionality, explained in
    Lemma 1 in the paper to be equivalent under certain conditions. 
    """
    cnf = []
    for p in progressbar(list(allProfiles())):
        for a in allAlternatives():
            if isSemiPartylistProfile(p, a):
                for k in ks:
                    if p.count(singletonBallot(a)) >= n/k-1e-5:
                        for c in allCommitteesOfSize(k):
                            if not c[a]:
                                cnf.append([negLiteral(c, p)])

    return cnf

def cnfWeakEfficiency():
    """
    If at least k candidates get a vote, don't elect someone without votes.
    """
    cnf = []
    for p in progressbar(list(allProfiles())):
        for k in ks:
            un = union(p)
            if sum(un) >= k:
                for c in allCommitteesOfSize(k):
                    if not isSubsetOf(c, un, strict=False):
                        cnf.append([negLiteral(c, p)])
    return cnf

def cnfJustifiedRepresentation():
    """
    Every committee elected must satisfy Justified Representation
    """
    cnf = []
    for p in progressbar(list(allProfiles())):
        for c in allCommittees():
            if not satisfiesJR(c, p):
                cnf.append([negLiteral(c,p)])
    return cnf


def cnfExtendedJustifiedRepresentation():
    """
    Every committee elected must satisfy Justified Representation
    """
    cnf = []
    for p in progressbar(list(allProfiles())):
        for c in allCommittees():
            if not satisfiesEJR(c, p):
                cnf.append([negLiteral(c,p)])
    return cnf


def cnfAtLeastOne():
    """
    Every profile should give at least a single winning committee 
    (per definition of a voting rule).
    """
    cnf = []
    for r in list(allProfiles()):
        for k in ks:
            clause = []
            for c in allCommitteesOfSize(k):
                clause.append(posLiteral(c, r))
            cnf.append(clause)
    return cnf


def cnfResolute():
    """
    Every profile should give exactly one winning committee.
    """
    cnf = []
    for r in progressbar(list(allProfiles())):
        for k in ks:
            for c1,c2 in combinations(allCommitteesOfSize(k), 2):
                cnf.append([negLiteral(c1, r), negLiteral(c2, r)])
    return cnf



# Running this at the start before any other cnf generator ensures that the literals
# always correspond to the same ballot/outcome combination, allowing for consistency
# in any downstream application
cnfAtLeastOne()

if __name__ == '__main__':

    # collect a dictionary of all available axioms
    cnfFunctions = [x for x in dir() if x.startswith("cnf") and x != "cnfAtLeastOne"]
    cnfFunctionsDict = {x[3:].lower(): eval(x) for x in cnfFunctions}

    parser = argparse.ArgumentParser()
    parser.add_argument("axioms", nargs='*',
                        help="the list of axioms, separated by spaces, to include in the cnf")
    parser.add_argument("-l", "--list", action="store_true",
                        help="print a list of all implemented axioms and exit")
    parser.add_argument("-o", "--outputfile", default="output.dimacs",
            help="the name of the file in which the cnf will be output (default: 'output.dimacs')")
    parser.add_argument("-s", "--satsolve", action="store_true",
            help="check whether the generated cnf is satisfiable (note: requires the dependency 'pylgl')")
    parser.add_argument("-p", "--progressbar", action="store_true",
            help="show a progress bar during the generation of the axioms (note: requires the dependency 'tqdm')")

    args = parser.parse_args()

    if args.list:
        # Print all the available axioms
        print("The following axioms are available:")
        print('\n'.join([x[3:] for x in cnfFunctions]))
        exit()

    if args.progressbar:
        import tqdm
        SHOW_PROGRESS = True

    if args.satsolve:
        import pylgl

    for axiomName in args.axioms:
        if axiomName.lower() not in cnfFunctionsDict:
           print("'{}' is not an implemented axiom".format(axiomName), file=stderr)
           exit()

    # generate the cnf
    cnf = []
    cnf += cnfAtLeastOne()
    for num, axiomName in enumerate(args.axioms):
        print("Now generating {}/{}: {}".format(num+1, len(args.axioms), axiomName))
        cnf += cnfFunctionsDict[axiomName.lower()]()

    # output the cnf in DIMACS format
    print("Writing output to '{}'".format(args.outputfile))
    dimacs(cnf, litcount, args.outputfile)

    if args.satsolve:
        result = pylgl.solve(cnf)
        if result == 'UNSAT':
            print("Unsatisfiable")
        else:
            print("Satisfiable")
